/* libMath: Mathematical library

 Bugs can be reported to yokochi@pharm.hokudai.ac.jp

 This work is a part of Olivia development project.
   http://fermi.pharm.hokudai.ac.jp/olivia/
*/

#ifndef __LIB_MATH__
#define __LIB_MATH__

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <math.h>
#include <assert.h>

#ifndef __OLIVIA_IRIX__

#include <stdint.h>

#else

#include <alloca.h>

#endif

#ifdef __XYZA2PIPE__

typedef struct {
 float r, i;
} complex;

typedef struct {
 double r, i;
} doublecomplex;

#else

#include <f2c.h>

#endif

typedef struct {
 complex x, y;
} hypercomplex;

typedef struct {
 doublecomplex x, y;
} doublehypercomplex;

#if defined(__linux__) && defined(__USE_ISOC99) || defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__) || defined(__APPLE__) || defined(__CYGWIN__) || defined(__INTERIX)

#ifndef HAVE_C99MATHEXT
#define HAVE_C99MATHEXT
#endif

#endif

/* C99 <math.h> extension */

#ifndef HAVE_C99MATHEXT

#define isnormal(x) (sizeof(x)==sizeof(float)?__isnormalf((float)(x)):__isnormal((double)(x)))

int __isnormalf(float x);
int __isnormal(double x);

#define isfinite(x) (sizeof(x)==sizeof(float)?__isfinitef((float)(x)):__isfinite((double)(x)))

int __isfinitef(float x);
int __isfinite(double x);

#endif

#define powi(x,y) (sizeof(x)==sizeof(float)?__powif((float)(x),(int)(y)):__powi((double)(x),(int)(y)))

float __powif(float x, int n);
double __powi(double x, int n);

/* complex */

complex Cadd(complex a, complex b);
complex Csub(complex a, complex b);
complex Cmul(complex a, complex b);
complex Conj(complex z);
complex Cdiv(complex a, complex b);
float Cabs(complex z);
complex Csqrt(complex z);
complex RCmul(float x, complex a);
complex Cexp(complex z);

/* doublecomplex */

doublecomplex DCadd(doublecomplex a, doublecomplex b);
doublecomplex DCsub(doublecomplex a, doublecomplex b);
doublecomplex DCmul(doublecomplex a, doublecomplex b);
doublecomplex DConj(doublecomplex z);
doublecomplex DCdiv(doublecomplex a, doublecomplex b);
double DCabs(doublecomplex z);
doublecomplex DCsqrt(doublecomplex z);
doublecomplex RDCmul(double x, doublecomplex a);
doublecomplex DCexp(doublecomplex z);

/* hypercomplex */

hypercomplex HCadd(hypercomplex a, hypercomplex b);
hypercomplex HCsub(hypercomplex a, hypercomplex b);
hypercomplex HCmul(hypercomplex a, hypercomplex b);
hypercomplex HConj(hypercomplex z);
hypercomplex HCtp(hypercomplex z);

#endif
