/* libMatrix: Memory allocation/free library

 Bugs can be reported to yokochi@pharm.hokudai.ac.jp

 This work is a part of Olivia development project.
   http://fermi.pharm.hokudai.ac.jp/olivia/
*/

#include "libMatrix.h"

float *fmalloc1d(const size_t size1)
{
 float *p;

 assert(p = (float *) malloc(size1 * sizeof(float)));

 return p;
}

float **fmalloc2d(const size_t size1, const size_t size2)
{
 int i;
 float **p;

 assert(p = (float **) malloc(size1 * sizeof(float *)));

 assert(p[0] = (float *) malloc(size1 * size2 * sizeof(float)));

 for (i = 1; i < size1; i++)
  p[i] = p[i - 1] + size2;

 return p;
}

float ***fmalloc3d(const size_t size1, const size_t size2, const size_t size3)
{
 int i, j;
 float ***p;

 assert(p = (float ***) malloc(size1 * sizeof(float **)));

 assert(p[0] = (float **) malloc(size1 * size2 * sizeof(float *)));

 assert(p[0][0] = (float *) malloc(size1 * size2 * size3 * sizeof(float)));

 for (j = 1; j < size2; j++)
  p[0][j] = p[0][j - 1] + size3;

 for (i = 1; i < size1; i++) {
  p[i] = p[i - 1] + size2;
  p[i][0] = p[i - 1][0] + size2 * size3;
  for (j = 1; j < size2; j++)
   p[i][j] = p[i][j - 1] + size3;
 }

 return p;
}

float ****fmalloc4d(const size_t size1, const size_t size2, const size_t size3, const size_t size4)
{
 int i, j, k;
 float ****p;

 assert(p = (float ****) malloc(size1 * sizeof(float ***)));

 assert(p[0] = (float ***) malloc(size1 * size2 * sizeof(float **)));

 assert(p[0][0] = (float **) malloc(size1 * size2 * size3 * sizeof(float *)));

 assert(p[0][0][0] = (float *) malloc(size1 * size2 * size3 * size4 * sizeof(float)));

 for (k = 1; k < size3; k++)
  p[0][0][k] = p[0][0][k - 1] + size4;

 for (j = 1; j < size2; j++) {
  p[0][j] = p[0][j - 1] + size3;
  p[0][j][0] = p[0][j - 1][0] + size3 * size4;
  for (k = 1; k < size3; k++)
   p[0][j][k] = p[0][j][k - 1] + size4;
 }

 for (i = 1; i < size1; i++) {
  p[i] = p[i - 1] + size2;
  p[i][0] = p[i - 1][0] + size2 * size3;
  p[i][0][0] = p[i - 1][0][0] + size2 * size3 * size4;

  for (k = 1; k < size3; k++)
   p[i][0][k] = p[i][0][k - 1] + size4;

  for (j = 1; j < size2; j++) {
   p[i][j] = p[i][j - 1] + size3;
   p[i][j][0] = p[i][j - 1][0] + size3 * size4;
  }
 }

 for (i = 1; i < size1; i++) {
  for (k = 1; k < size3; k++)
   p[i][0][k] = p[i - 1][0][k - 1] + size2 * size3 * size4 + size4;

  for (j = 1; j < size2; j++) {
   p[i][j] = p[i - 1][j - 1] + size2 * size3 + size3;
   p[i][j][0] = p[i - 1][j - 1][0] + size2 * size3 * size4 + size3 * size4;
   for (k = 1; k < size3; k++)
    p[i][j][k] = p[i - 1][j][k - 1] + size2 * size3 * size4 + size4;
  }
 }

 return p;
}

void free_fmatrix1d(float *p)
{
 if (p == NULL)
  return;

 free((char *) (p));

 p = NULL;
}

void free_fmatrix2d(float **p)
{
 if (p == NULL)
  return;

 free((char *) (p[0]));
 free((char *) (p));

 p = NULL;
}

void free_fmatrix3d(float ***p)
{
 if (p == NULL)
  return;

 free((char *) (p[0][0]));
 free((char *) (p[0]));
 free((char *) (p));

 p = NULL;
}

void free_fmatrix4d(float ****p)
{
 if (p == NULL)
  return;

 free((char *) (p[0][0][0]));
 free((char *) (p[0][0]));
 free((char *) (p[0]));
 free((char *) (p));

 p = NULL;
}
