/* xyza2pipe: Cross conversion environment of NMR spectra

 Bugs can be reported to yokochi@pharm.hokudai.ac.jp

 This program is a part of Olivia development project.
   http://fermi.pharm.hokudai.ac.jp/olivia/
*/

#include "xyza2pipe.h"

char header[PIPE_HEADERSIZE];
char axislabel[4][MAXASSNAME + 1] = { {""} };

char axisname[4][MAXASSNAME + 1];

char dimension, byteswap;

char swapdata = 0, swappar = 0, usrlabel = 0, usrshift = 0;
char leftcar = 0, extleft = 0, adjcar = 0, adjh2o = 0, relyof = 0;
char usrphase[4] = { 0 };

short headersize;
short blocksize[4];

int datasize[4], datasize_orig[4];
int unitsize[4];

float obsfreq[4];
float spcenter[4];
float origfreq[4];
float spwidth[4];
float usrcenter[4] = { NULLPPM };

float phase[4][2];

const float PIPE_HEADER[3] = { PIPE_HEADER_0, PIPE_HEADER_1, PIPE_HEADER_2 };

const char *usage[] = {
 "pipe2azara -o/--out outTemplate(2D/3D/4D) < stdin\n", "Optional Arguments:\n", " -s/--swap  Byteswap data\n",
     " -l/--left  Left shift observable carrier frequency by sw/2\n",
     " -g/--orig  Rely on the origin frequency for spectral regulation\n", " --xLAB X-Axis Label\n",
     " --yLAB Y-Axis Label\n", " --zLAB Z-Axis Label\n", " --aLAB A-Axis Label\n", " --xCAR X-Axis Center [ppm]\n",
     " --yCAR Y-Axis Center [ppm]\n", " --zCAR Z-Axis Center [ppm]\n", " --aCAR A-Axis Center [ppm]\n\n", ""
};

int main(int argc, char *argv[])
{
 char axis_option = 'x';
 char filename[MAXLONGNAME] = { "" };
 int l, opt_idx = 0;

 static struct option options[] = {
  {"out", required_argument, 0, 'o'},
  {"swap", no_argument, 0, 's'},
  {"left", no_argument, 0, 'l'},
  {"orig", no_argument, 0, 'g'},
  {"xLAB", required_argument, 0, 6},
  {"yLAB", required_argument, 0, 7},
  {"zLAB", required_argument, 0, 8},
  {"aLAB", required_argument, 0, 9},
  {"xCAR", required_argument, 0, 10},
  {"yCAR", required_argument, 0, 11},
  {"zCAR", required_argument, 0, 12},
  {"aCAR", required_argument, 0, 13},
  {0}
 };

 while (1) {
  l = getopt_long(argc, argv, "o:slg", options, &opt_idx);

  if (l == -1)
   break;

  switch (l) {
  case 0:
   if (options[opt_idx].flag != 0)
    break;
   fprintf(stderr, "option %s", options[opt_idx].name);
   if (optarg)
    fprintf(stderr, " with arg %s", optarg);
   fputc('\n', stderr);
   break;
  case 'o':
   if (optarg)
    strncpy(filename, optarg, MAXLONGNAME);
   break;
  case 's':
   swapdata = 1;
   break;
  case 'l':
   leftcar = 1;
   break;
  case 'g':
   relyof = 1;
   break;
  case 6:
   if (optarg) {
    usrlabel = 1;
    strncpy(axislabel[0], optarg, MAXASSNAME);
   }
   break;
  case 7:
   if (optarg) {
    usrlabel = 1;
    strncpy(axislabel[1], optarg, MAXASSNAME);
   }
   break;
  case 8:
   if (optarg) {
    usrlabel = 1;
    strncpy(axislabel[2], optarg, MAXASSNAME);
   }
   break;
  case 9:
   if (optarg) {
    usrlabel = 1;
    strncpy(axislabel[3], optarg, MAXASSNAME);
   }
   break;
  case 10:
   usrshift = 1;
   usrcenter[0] = atof(optarg);
   break;
  case 11:
   usrshift = 1;
   usrcenter[1] = atof(optarg);
   break;
  case 12:
   usrshift = 1;
   usrcenter[2] = atof(optarg);
   break;
  case 13:
   usrshift = 1;
   usrcenter[3] = atof(optarg);
   break;
  case '?':			/* getopt_long already printed an error message. */
   break;
  }
 }

 if (argc == 2)
  strncpy(filename, argv[1], MAXLONGNAME);

 else if (optind < argc) {
  fprintf(stderr, "non-option ARGV-elements: ");
  while (optind < argc)
   fprintf(stderr, "%s ", argv[optind++]);
  fputc('\n', stderr);

  return 1;
 }

 if (filename[0] == 0) {
  l = 0;
  while (strcmp(usage[l], "") != 0)
   fprintf(stderr, "%s", usage[l++]);

  return 1;
 }

 if (checkpipe() != 0)
  return 1;

 cnvhdr(axis_option, 'b');

 switch (dimension) {
 case 2:
  if (pullazara2d(filename) != 0)
   return 1;
  break;

 case 3:
  if (pullazara3d(filename) != 0)
   return 1;
  break;

 case 4:
  if (pullazara4d(filename) != 0)
   return 1;
  break;
 }

 return 0;
}
